from datetime import datetime
import unittest

from threatbus.data import Intel, IntelData, IntelType, Operation, Sighting
from threatbus_misp.message_mapping import map_to_internal, map_to_misp


class TestMessageMapping(unittest.TestCase):
    def setUp(self):
        self.raw_ts = 1579104545
        self.ts = datetime.fromtimestamp(self.raw_ts)
        self.id = "15"
        self.indicator = "example.com"
        self.expected_intel_type = IntelType.DOMAIN
        self.valid_misp_attribute = {
            "id": self.id,
            "event_id": "1",
            "object_id": "0",
            "object_relation": None,
            "category": "Network activity",
            "type": "domain",
            "value1": self.indicator,
            "value2": "",
            "to_ids": True,
            "uuid": "5e1f2787-fcfc-4718-a58a-00b4c0a82f06",
            "timestamp": str(self.raw_ts),
            "distribution": "5",
            "sharing_group_id": "0",
            "comment": "",
            "deleted": False,
            "disable_correlation": False,
            "value": self.indicator,
            "Sighting": [],
        }

    def test_invalid_inputs(self):
        self.assertIsNone(map_to_misp(None))
        self.assertIsNone(map_to_misp("Hello"))
        self.assertIsNone(map_to_misp(self))

    def test_invalid_misp_inputs(self):
        self.assertIsNone(map_to_internal(None, None))
        self.assertIsNone(map_to_internal(None, "delete"))
        self.assertIsNone(map_to_internal(self.valid_misp_attribute, None))

    def test_default_action_remove(self):
        self.assertIsNotNone(map_to_internal(self.valid_misp_attribute, "FOOOO"))

    def test_valid_misp_inputs(self):
        intel_data = IntelData(self.indicator, self.expected_intel_type, source="MISP")
        expected_intel = Intel(self.ts, self.id, intel_data, Operation.ADD)
        self.assertEqual(
            map_to_internal(self.valid_misp_attribute, "add"), expected_intel
        )

    def test_valid_inputs(self):
        sighting = Sighting(self.ts, self.id, {})
        self.assertIsNotNone(map_to_misp(sighting))



from setuptools import setup
import pathlib

plugin_dir = pathlib.Path(__file__).parent.absolute()

with open(f"{plugin_dir}/README.md", "r") as fh:
    long_description = fh.read()

setup(
    author="Tenzir",
    author_email="engineering@tenzir.com",
    classifiers=[
        # https://pypi.org/classifiers/
        "Development Status :: 3 - Alpha",
        "Environment :: Plugins",
        "License :: OSI Approved :: BSD License",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Interface Engine/Protocol Translator",
        "Topic :: Security",
        "Topic :: Software Development :: Object Brokering",
        "Topic :: System :: Distributed Computing",
    ],
    description="A plugin to enable threatbus communication with MISP.",
    entry_points={"threatbus.app": ["misp = threatbus_misp.plugin"]},
    install_requires=[
        "threatbus>=2020.04.29",
        "pymisp>=2.4.120",
        "pyzmq>=18.1.1",
        "confluent-kafka>=1.3.0",
    ],
    keywords=[
        "threatbus",
        "MISP",
        "threat intelligence",
        "IDS",
        "zeromq",
        "zmq",
        "kafka",
    ],
    license="BSD 3-clause",
    long_description=long_description,
    long_description_content_type="text/markdown",
    name="threatbus-misp",
    package_dir={"": "plugins/apps"},
    packages=["threatbus_misp"],
    python_requires=">=3.7",
    url="https://github.com/tenzir/threatbus",
    version="2020.07.28",
)



from datetime import datetime
from threatbus.data import Intel, IntelData, IntelType, Operation, Sighting
import pymisp

misp_intel_type_mapping = {
    "ip-src": IntelType.IPSRC,
    "ip-dst": IntelType.IPDST,
    "ip-src|port": IntelType.IPSRC_PORT,
    "ip-dst|port": IntelType.IPDST_PORT,
    "email-src": IntelType.EMAILSRC,
    "email-dst": IntelType.EMAILDST,
    "target-email": IntelType.TARGETEMAIL,
    "email-attachment": IntelType.EMAILATTACHMENT,
    "filename": IntelType.FILENAME,
    "hostname": IntelType.HOSTNAME,
    "domain": IntelType.DOMAIN,
    "domain|ip": IntelType.DOMAIN_IP,
    "url": IntelType.URL,
    "user-agent": IntelType.USERAGENT,
    "md5": IntelType.MD5,
    "malware-sample": IntelType.MALWARESAMPLE,
    "filename|md5": IntelType.FILENAME_MD5,
    "sha1": IntelType.SHA1,
    "filename|sha1": IntelType.FILENAME_SHA1,
    "sha256": IntelType.SHA256,
    "filename|sha256": IntelType.FILENAME_SHA256,
    "x509-fingerprint-sha1": IntelType.X509FINGERPRINTSHA1,
    "pdb": IntelType.PDB,
    "authentihash": IntelType.AUTHENTIHASH,
    "ssdeep": IntelType.SSDEEP,
    "imphash": IntelType.IMPHASH,
    "pehash": IntelType.PEHASH,
    "impfuzzy": IntelType.IMPFUZZY,
    "sha224": IntelType.SHA224,
    "sha384": IntelType.SHA384,
    "sha512": IntelType.SHA512,
    "sha512/224": IntelType.SHA512_224,
    "sha512/256": IntelType.SHA512_256,
    "tlsh": IntelType.TLSH,
    "cdhash": IntelType.CDHASH,
    "filename|authentihash": IntelType.FILENAME_AUTHENTIHASH,
    "filename|ssdeep": IntelType.FILENAME_SSDEEP,
    "filename|imphash": IntelType.FILENAME_IMPHASH,
    "filename|pehash": IntelType.FILENAME_PEHASH,
    "filename|impfuzzy": IntelType.FILENAME_IMPFUZZY,
    "filename|sha224": IntelType.FILENAME_SHA224,
    "filename|sha384": IntelType.FILENAME_SHA384,
    "filename|sha512": IntelType.FILENAME_SHA512,
    "filename|sha512/224": IntelType.FILENAME_SHA512_224,
    "filename|sha512/256": IntelType.FILENAME_SHA512_256,
    "filename|tlsh": IntelType.FILENAME_TLSH,
}


def map_to_internal(misp_attribute, action, logger=None):
    """Maps the given MISP attribute to the threatbus intel format.
        @param misp_attribute A MISP attribute
        @param action A string from MISP, describing the action for the attribute (either 'add' or 'delete')
        @return the mapped intel item or None
    """
    # parse MISP attribute
    if not misp_attribute:
        return None
    to_ids = misp_attribute.get("to_ids", False)
    if not to_ids and action != "edit" or not action:
        return None
    operation = Operation.REMOVE
    if (action == "edit" or action == "add") and to_ids:
        operation = Operation.ADD

    # parse values
    intel_type = misp_attribute.get("type", None)
    indicator = misp_attribute.get("value", None)
    if not intel_type or not indicator:
        return None

    # parse compound MISP intel:
    if "|" in intel_type:
        indicator = tuple(indicator.split("|"))
        if len(indicator) != 2:
            if logger:
                logger.debug(
                    f"Expected '|'-separated composite values for MISP intel type {intel_type}"
                )
            return None

    return Intel(
        datetime.fromtimestamp(int(misp_attribute["timestamp"])),
        str(misp_attribute["id"]),
        IntelData(indicator, misp_intel_type_mapping[intel_type], source="MISP",),
        operation,
    )


def map_to_misp(sighting):
    """Maps the threatbus sighting format to a MISP sighting.
        @param sighting A threatbus Sighting object
        @return the mapped MISP sighting object or None
    """
    if not sighting or not isinstance(sighting, Sighting):
        return None

    misp_sighting = pymisp.MISPSighting()
    misp_sighting.from_dict(
        id=sighting.intel,
        source=sighting.context.get("source", None),
        type="0",  # true positive sighting: https://www.circl.lu/doc/misp/automation/#post-sightingsadd
        timestamp=sighting.ts,
    )
    return misp_sighting



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    author="Tenzir",
    author_email="engineering@tenzir.com",
    classifiers=[
        # https://pypi.org/classifiers/
        "Development Status :: 3 - Alpha",
        "Environment :: Plugins",
        "License :: OSI Approved :: BSD License",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX :: Linux",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Interface Engine/Protocol Translator",
        "Topic :: Security",
        "Topic :: Software Development :: Object Brokering",
        "Topic :: System :: Distributed Computing",
    ],
    description="The missing link to connect open-source threat intelligence tools.",
    entry_points={"console_scripts": ["threatbus=threatbus.threatbus:main"]},
    include_package_data=True,
    install_requires=[
        "coloredlogs>=10.0",
        "confuse>=1.0",
        "pluggy>=0.13",
        "black>=19.10b",
    ],
    keywords=[
        "threatbus",
        "threat intelligence",
        "intel",
        "sightings",
        "open source threat intelligence",
    ],
    license="BSD 3-clause",
    long_description=long_description,
    long_description_content_type="text/markdown",
    name="threatbus",
    packages=["threatbus"],
    python_requires=">=3.7",
    setup_requires=["setuptools", "wheel"],
    url="https://github.com/tenzir/threatbus",
    version="2020.07.28",
)



from confluent_kafka import Consumer
from datetime import datetime
import json
import pymisp
from queue import Queue
import threading
import warnings
import zmq

from threatbus_misp.message_mapping import map_to_internal, map_to_misp
import threatbus
from threatbus.data import MessageType

warnings.simplefilter("ignore")  # pymisp produces urllib warnings

"""MISP - Open Source Threat Intelligence Platform - plugin for Threat Bus"""

plugin_name = "misp"
misp = None
lock = threading.Lock()


def validate_config(config):
    assert config, "config must not be None"
    # redact fallback values to allow for omitting configuration blocks
    config["api"].add({})
    config["zmq"].add({})
    config["kafka"].add({})
    if config["zmq"].get(dict) and config["kafka"].get(dict):
        raise AssertionError("either use ZeroMQ or Kafka, but not both")

    if config["api"].get(dict):
        config["api"]["host"].get(str)
        config["api"]["ssl"].get(bool)
        config["api"]["key"].get(str)
    if config["zmq"].get(dict):
        config["zmq"]["host"].get(str)
        config["zmq"]["port"].get(int)
    if config["kafka"].get(dict):
        config["kafka"]["topics"].get(list)
        config["kafka"]["poll_interval"].add(1.0)
        config["kafka"]["poll_interval"].get(float)
        config["kafka"]["config"].get(dict)


def publish_sightings(outq):
    """Reports / publishes true-positive sightings of intelligence items back to the given MISP endpoint.
        @param outq The queue from which to forward messages to MISP 
    """
    global logger, misp, lock
    if not misp:
        return
    while True:
        sighting = outq.get(block=True)
        logger.debug(
            f"report sighting for intel id {sighting.intel} seen at {sighting.ts}"
        )
        misp_sighting = map_to_misp(sighting)
        lock.acquire()
        misp.add_sighting(misp_sighting)
        lock.release()


def receive_kafka(kafka_config, inq):
    """Binds a Kafka consumer to the the given host/port. Forwards all received messages to the inq.
        @param kafka_config A configuration object for Kafka binding
        @param inq The queue to which intel items from MISP are forwarded to
    """
    consumer = Consumer(kafka_config["config"].get(dict))
    consumer.subscribe(kafka_config["topics"].get(list))
    global logger
    while True:
        message = consumer.poll(timeout=kafka_config["poll_interval"].get(float))
        if message is None:
            continue
        if message.error():
            logger.error(f"Kafka error: {message.error()}")
            continue
        try:
            msg = json.loads(message.value())
        except Exception as e:
            logger.error(f"Error decoding Kafka message: {e}")
            continue
        if not msg.get("Attribute", None):
            logger.debug("Skipping message without MISP Attribute")
            continue
        intel = map_to_internal(msg["Attribute"], msg.get("action", None), logger)
        if not intel:
            logger.debug(f"Discarding unparsable intel {msg['Attribute']}")
        else:
            inq.put(intel)


def receive_zmq(zmq_config, inq):
    """Binds a ZMQ poller to the the given host/port. Forwards all received messages to the inq.
        @param zmq_config A configuration object for ZeroMQ binding
        @param inq The queue to which intel items from MISP are forwarded to
    """

    socket = zmq.Context().socket(zmq.SUB)
    socket.connect(f"tcp://{zmq_config['host']}:{zmq_config['port']}")
    # TODO: allow reception of more topics, i.e. handle events.
    socket.setsockopt(zmq.SUBSCRIBE, b"misp_json_attribute")
    poller = zmq.Poller()
    poller.register(socket, zmq.POLLIN)

    while True:
        socks = dict(poller.poll(timeout=None))
        if socket in socks and socks[socket] == zmq.POLLIN:
            raw = socket.recv()
            _, message = raw.decode("utf-8").split(" ", 1)
            try:
                msg = json.loads(message)
            except Exception as e:
                logger.error(f"Erro decoding message {message}: {e}")
                continue
            if not msg.get("Attribute", None):
                continue
            intel = map_to_internal(msg["Attribute"], msg.get("action", None), logger)
            if not intel:
                logger.debug(f"Discarding unparsable intel {msg['Attribute']}")
            else:
                inq.put(intel)


@threatbus.app
def snapshot(snapshot_type, result_q, time_delta):
    global logger, misp, lock
    if snapshot_type != MessageType.INTEL:
        logger.debug(f"Sighting snapshot feature not yet implemented.")
        return  # TODO sighting snapshot not yet implemented

    if not misp:
        return
    logger.debug(f"Executing intel snapshot for time delta {time_delta}")
    lock.acquire()
    data = misp.search(
        controller="attributes", to_ids=True, date_from=datetime.now() - time_delta
    )
    lock.release()
    if not data:
        return
    for attr in data["Attribute"]:
        intel = map_to_internal(attr, "add", logger)
        if intel:
            result_q.put(intel)


@threatbus.app
def run(config, logging, inq, subscribe_callback, unsubscribe_callback):
    global logger
    logger = threatbus.logger.setup(logging, __name__)
    config = config[plugin_name]
    try:
        validate_config(config)
    except Exception as e:
        logger.fatal("Invalid config for plugin {}: {}".format(plugin_name, str(e)))
    if config["api"].get():
        host, key, ssl = (
            config["api"]["host"].get(),
            config["api"]["key"].get(),
            config["api"]["ssl"].get(),
        )
        try:
            global misp, lock
            lock.acquire()
            misp = pymisp.ExpandedPyMISP(url=host, key=key, ssl=ssl)
            lock.release()
        except Exception:
            # TODO: log individual error per MISP subscriber
            logger.error(f"Cannot subscribe to MISP at {host}, using SSL: {ssl}")
            lock.release()

    # TODO: MISP instances shall subscribe themselves to threatbus and each subscription shall have an individual outq and receiving thread for intel updates.
    outq = Queue()
    subscribe_callback("threatbus/sighting", outq)

    if misp:
        threading.Thread(target=publish_sightings, args=(outq,), daemon=True).start()
    if config["zmq"].get():
        threading.Thread(
            target=receive_zmq, args=(config["zmq"], inq), daemon=True
        ).start()
    if config["kafka"].get():
        threading.Thread(
            target=receive_kafka, args=(config["kafka"], inq), daemon=True
        ).start()
    logger.info("MISP plugin started")
